#!/bin/bash
########################################################################
# build_hepmc3.sh
#
# Build hepmc3 and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=hepmc3
origpkgver=v3_2_1
pkgver=${origpkgver}b # Append letters for local patches.
ssibuildshims_version=v2_03_02
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=HepMC3-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e17,e19,e20,e21,c2,c7} \
  --pqual={p2715a,p372,p383a} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_BUILD

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  WANT_TESTS

    Activate tests.

EOF
##################

# Needs to be early so we can give a useful error message rather than
# having the UPS setup attempt fail.
if [[ "${pqual}" =~ ^p([[:digit:]]+) ]] && \
  { (( ${BASH_REMATCH[1]:0:1} == 2 )) || (( ${BASH_REMATCH[1]:1} < 83 )); } && \
  { { [ "${cqual:0:1}" = e ] && (( ${cqual#e} > 19 )); } || \
    { [ "${cqual:0:1}" = c ] && (( ${cqual#c} > 7 )); } || \
    ! [[ "${cqual:0:1}" == [ce] ]]; }; then
  case ${cqual:0:1} in
    e) emsg="GCC > 8.3";;
    c) emsg="LLVM/Clang > 7.0.0";;
    *) emsg="non-GCC / LLVM compilers"
  esac
  ssi_die "Python < 3.8.3 not supported with ${emsg}"
fi

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS FQ_TABLE_DIR || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*

# Require CMake.
${trace} setup_cmake --latest  || exit

# We can use GNU Make, but we'll take Ninja if it's available
${trace} setup_ninja --optional && (( want_ninja = 1 ))

# Valgrind's memcheck is used for tests.
setup_prod valgrind --current --optional
valgrind=$(which valgrind 2>/dev/null)
[ -n "${valgrind}" ] && ssi_info "using valgrind at ${valgrind}"

####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  fix-HepMC3-config.patch \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cmake_options
(( want_ninja )) && cmake_options+=(-GNinja)

case ${cqual} in
  e1[79]|e20) cc=gcc; cxx=g++; cxxstd=17;;
  e21) cc=gcc; cxx=g++; cxxstd=20;;
  c[27]) cc=clang; cxx=clang++; cxxstd=17;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac
case ${btype} in
  debug) cmake_build_type=Debug;;
  prof) cmake_build_type=RelWithDebInfo;;
  *) ssi_die -I "accepted unknown btype=${btype}"
esac
case ${pqual} in
  p2715a|p372|p383a) ${trace} python_vars || exit
    absolute_path -- "${blddir}/${PYTHON_LIBDIR}/site-packages" tmp_sitearch;;
  *) ssi_die -I "accepted unknown pqual=${pqual}"
esac

cmake_options+=( \
  -DCMAKE_INSTALL_PREFIX="${pkgdir}" \
  -DCMAKE_BUILD_TYPE="${cmake_build_type}" \
  -DHEPMC3_ENABLE_SEARCH:BOOL=OFF \
  -DHEPMC3_ENABLE_ROOTIO:BOOL=OFF \
  -DHEPMC3_ENABLE_PYTHON:BOOL=ON \
  -DHEPMC3_PYTHON_VERSIONS="${pythonMdotm}" \
  -DHEPMC3_Python_SITEARCH${pythonMm}="${tmp_sitearch}" \
  -DHEPMC3_BUILD_DOCS:BOOL=OFF \
  -DHEPMC3_BUILD_EXAMPLES:BOOL=OFF \
  -DHEPMC3_INSTALL_INTERFACES:BOOL=ON \
  -DHEPMC3_ENABLE_TEST:BOOL=ON \
  -DCMAKE_C_COMPILER=${cc} \
  -DCMAKE_C_FLAGS_RELEASE="-O3 -DNDEBUG" \
  -DCMAKE_C_FLAGS_RELWITHDEBINFO="-g -O3 -DNDEBUG -fno-omit-frame-pointer" \
  -DCMAKE_C_FLAGS_DEBUG="-g -O0" \
  -DCMAKE_CXX_COMPILER=${cxx} \
  -DCMAKE_CXX_FLAGS_RELEASE="-O3 -DNDEBUG" \
  -DCMAKE_CXX_FLAGS_RELWITHDEBINFO="-g -O3 -DNDEBUG -fno-omit-frame-pointer -Wall -Wextra -pedantic" \
  -DCMAKE_CXX_FLAGS_DEBUG="-g -O0" \
  -DCMAKE_CXX_STANDARD=${cxxstd} \
  -DCMAKE_CXX_EXTENSIONS=OFF \
  )

##################
# Ascertain and remove packages we don't like from HEPMC3_TEST_PACKAGES
# to avoid accidentally finding something.
disabled_test_pkgs=(Pythia8 Photospp Tauolapp MCTester HepMC2)
ssi_info -i "disabling unwanted external packages used by optional tests:\n" \
  "  ${disabled_test_pkgs[*]}"

test_pkgs=($(${trace} sed -nEe 's&^\s*(set|SET)\s*\(HEPMC3_TEST_PACKAGES\s+"([^\"]+)".*&\2&
t found; d; : found {; s&\s*,\s*& &g; p; q; };' \
  "${sourcedir}/${pkgtarname}/test/CMakeLists.txt"))
enabled_test_pkgs=
for pkg in "${test_pkgs[@]}"; do
  item_in ${pkg} "${disabled_test_pkgs[@]}" || \
    enabled_test_pkgs+="${enabled_test_pkgs:+;}${pkg}"
done
if [ -n  "${enabled_test_pkgs}" ]; then
  cmake_options+=(-DHEPMC3_TEST_PACKAGES="${enabled_test_pkgs}")
else
  # Something went wrong (CMake file change?): backstop.
  ssi_warn -i "unable to obtain test packages list from " \
    "${sourcedir}/${pkgtarname}/test/CMakeLists.txt\n" \
    "Disabling find_package() calls as a fallback"
  for pkg in "${disabled_test_pkgs[@]}"; do
    # MCTester is automatically disabled if ROOTIO is not enabled
    # (above).
    [ "${pkg}" = MCTester ] && continue
    cmake_options+=(-DCMAKE_DISABLE_FIND_PACKAGE_${pkg}:BOOL=ON)
  done
fi
##################

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

# For ctest:
[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

${trace} cmake "${cmake_options[@]}" "${sourcedir}/${pkgtarname}" || \
ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
ssi_die "build failed"
##################

##################
# Test. *C*

(( WANT_TESTS )) && { ${trace} ctest || ssi_die "tests failed"; }
##################

##################
# Install. *C*

${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

# Ascertain the actual libdir so we can fill in the arch-specific table
# file.
${trace} verify_in_filesystem -d --silent --var libdir -- \
  "${pkgdir}/lib"{64,} || \
  ssi_die "unable to ascertain LIBDIR from CMake cache information"

absolute_path -- "${libdir}/${PYTHON_LIBDIR}/site-packages" sitearch_final

${trace} verify_in_filesystem -d --silent --var sitearch -- \
  "${sitearch_final}" "${tmp_sitearch}" || \
  ssi_die "python bindings not installed in expected places:\n" \
  "  ${sitearch_final} or\n  ${tmp_sitearch}"

if ! [ "${sitearch}" = "${sitearch_final}" ]; then
  ssi_info "Moving Python bindings to final location ${sitearch_final}"
  ${trace} mv -fv "${sitearch%/*}" "${libdir}/" && \
    [ -d "${sitearch_final}" ] || \
    ssi_die "unable to relocate Python bindings from\n" \
    "  ${sitearch} -> ${sitearch_final}"
fi

ssi_info "Creating architecture-specific UPS table file due to GNUInstallDirs"
${trace} mkdir -p "${pkgups}" || \
  ssi_die "unable to create arch-specific UPS table directory ${pkgups}"

${trace} perlfilt -wpe 's&\@LIBDIR\@&'"${libdir##*/}"'&g' "${pkgtop}/ups/${package}.table" > \
  "${pkgups}/${package}.table" && \
  [[ -s "${pkgups}/${package}.table" ]] || \
  ssi_die -i "unable to create arch-specific UPS table file\n" \
  "  ${pkgups}/${package}.table"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*

ssi_info "post-processing for relocatability"
${trace} fix_config_script "${pkgdir}/bin/HepMC3-config" || \
  ssi_die "post-processing failed"
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
